__version__ = "0.0.12"
# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00__init__.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/00__init__.ipynb 3
from itertools import chain
from ptyp import (
    type, util, grds, meta, ptyp,
    # ptyp as amod,    
)

from .type import *
from .util import *
from .grds import *
from .meta import *
from .ptyp import *

# %% ../nbs/00__init__.ipynb 4
__all__ = sorted(set(chain(*[
    mod.__all__ for mod in (type, util, grds, meta, ptyp,)
])))



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/01_util.ipynb.

# %% auto 0
__all__ = ['optargname', 'args2strs', 'typescsv', 'typesusv', 'types2opts', 'guardtype', 'guardname', 'guarddocs', 'guardanns',
           'setguardattrs', 'classname', 'qualname', 'settypes']

# %% ../nbs/01_util.ipynb 6
import os, inspect
from functools import wraps

# %% ../nbs/01_util.ipynb 8
from typing import (Union, Callable, Optional, TypeGuard,)

# %% ../nbs/01_util.ipynb 10
#| export


# %% ../nbs/01_util.ipynb 12
#| export


# %% ../nbs/01_util.ipynb 14
from .type import (T, P, Types, Guard, GuardFunc, StrFactory, StrBool)

# %% ../nbs/01_util.ipynb 17
def optargname(arg: T) -> str:
    '''Gets the name of the (possibly `Optional`) argument.'''
    name = str(getattr(arg, '__name__', arg))
    if name == 'Optional' and len(args := getattr(arg, '__args__', None)) > 0:
        name = str(getattr(args[0], '__name__', args[0]))
    return name

def args2strs(*args) -> list[str]:
    '''Converts a sequence of arguments into a list of their type names as strings.'''
    astrs = [str(optargname(arg)) for arg in args or tuple()]
    return astrs

def typescsv(*types: Types) -> str:
    '''Generates a comma-separated string of type names.'''
    return ', '.join(args2strs(*types))

def typesusv(*types: Types) -> str:
    '''Generates an underscore-separated string of type names.'''
    return '_'.join(args2strs(*types))

# %% ../nbs/01_util.ipynb 19
def types2opts(*types: Types) -> Types:
    '''Converts a tuple of types into a tuple of `Optional` types, if `None` is included.'''
    if None in types: 
        types = tuple(Optional[t] for t in types if t is not None) 
    return types


def guardtype(*types: Types) -> Guard:
    '''Creates a TypeGuard based on the provided types.
    
    Parameters
    ----------
    types : Types
        A tuple of types to be included in the TypeGuard.
    
    Returns
    -------
    Guard
        A TypeGuard that represents either None (if no types are provided),
        a single type (if one type is provided), or a Union of the provided types.
    '''
    types = types2opts(*types)
    if len(types) == 0:  return TypeGuard[None]
    elif len(types) == 1: return TypeGuard[types[0]]
    else: return TypeGuard[Union[types[0], *types[1:]]]


# %% ../nbs/01_util.ipynb 21
def guardname(*types: Types, negate: bool = False) -> str:
    '''Generates a name for a guard function based on the provided types and negation.'''
    isstr = 'is' if not negate else 'not'
    return f'{isstr}{typesusv(*types).lower().replace("_", "")}'

def guarddocs(*types: Types, negate: bool = False) -> str: 
    '''Generates a docstring for a guard function based on the provided types and negation.'''
    isstr = 'is' if not negate else 'is not'
    adj = 'one of: ' if len(types) > 1 else ''
    return f'Check if `x` {isstr} {adj}{typescsv(*types)}.'

def guardanns(*types: Types, negate: bool = None) -> dict:
    '''Generates a dictionary of annotations for a guard function.'''
    return {'x': 'typing.Any', 'return': guardtype(*types)}

def setguardattrs(
    call: GuardFunc, 
    *types: Types,
    docsfn: StrFactory = guarddocs,
    namefn: StrFactory = guardname,
    annsfn: StrFactory = guardanns,
    negate: bool = False,
) -> Callable[P, TypeGuard[T]]:
    '''Sets the guard function's attributes including docstring, name, and annotations.

    Parameters
    ----------
    call : GuardFunc
        The guard function to be modified.
    types : Types
        A tuple of types for the guard.
    docsfn : StrFactory, optional
        The function to generate the docstring, by default guarddocs.
    namefn : StrFactory, optional
        The function to generate the name, by default guardname.
    annsfn : StrFactory, optional
        The function to generate the annotations, by default guardanns.
    negate : bool, optional
        A flag to indicate if the guard should represent a negation, by default False.

    Returns
    -------
    Callable[P, TypeGuard[T]]
        The modified guard function.
    '''
    # Update docstring, name, and annotations
    for fn, at in zip((docsfn, namefn, annsfn), ('__doc__', '__name__', '__annotations__')):
        attr = fn(*types, negate=negate)
        setattr(call, at, getattr(call, at, attr))
    return call

# %% ../nbs/01_util.ipynb 23
def classname(cls: type, clsname: StrBool = True) -> StrBool:
    '''Retrieves an optional classname for the given `cls`. 
    
    Notes
    -----
    - if `clsname == False`, then `False` is returned.
    - if `clsname` is a `str`, then that string is returned.
    - if `clsname` is a class itself, then the name of the class is returned.
    - if `clsname == True` or otherwise, then the name of the `cls` is returned. 
    '''
    if clsname == False: return clsname
    if inspect.isclass(clsname): return clsname.__name__
    elif isinstance(clsname, str): return clsname
    else: return cls.__name__
    
def qualname(
    cls, *types: Types, clsname: StrBool = True, 
    prefix: Optional[str] = None, suffix: Optional[str] = None, dropnone: bool = True
) -> str:
    '''Generates a qualified name for a type or a set of types.

    Parameters
    ----------
    cls : type
        The class for which to generate the name.
    types : Types
        A tuple of types.
    clsname : StrBool, optional
        Determines the format of the class name returned, by default True.
    prefix : Optional[str], optional
        An optional prefix for the name, by default None.
    suffix : Optional[str], optional
        An optional suffix for the name, by default None.
    dropnone : bool, optional
        Flag to drop None from types, by default True.

    Returns
    -------
    str
        The qualified name generated.
    '''
    if dropnone:
        types = tuple(Optional[t] for t in types if t is not None) 
    tname = typescsv(*types)
    qname = f'{{{tname}}}'
    if prefix is not None: qname = f'{prefix}{qname}'
    if suffix is not None: qname = f'{qname}{suffix}'
    if (cname := classname(cls, clsname)): qname = f'{cname}({qname})'
    return qname

def settypes(
    cls, *types: Types, clsname: StrBool = True,
    prefix: Optional[str] = None, suffix: Optional[str] = None, dropnone: bool = True,
    **kwargs
):
    '''Sets type-related attributes on a class.

    Parameters
    ----------
    cls : type
        The class to modify.
    types : Types
        A tuple of types.
    clsname : StrBool, optional
        Determines the format of the class name returned, by default True.
    prefix : Optional[str], optional
        An optional prefix for the name, by default None.
    suffix : Optional[str], optional
        An optional suffix for the name, by default None.
    dropnone : bool, optional
        Flag to drop None from types, by default True.

    Returns
    -------
    None
    '''
    qname = qualname(cls, *types, clsname=clsname, prefix=prefix, suffix=suffix, dropnone=dropnone)
    setattr(cls, '__qualname__', qname)
    annots = getattr(getattr(cls, 'check', None), '__annotations__', {})    
    annots['return'] = guardtype(*types)
    setattr(cls, 'types', types)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/02_grds.ipynb.

# %% auto 0
__all__ = ['guard', 'tset', 'tlist', 'tdict', 'ttype', 'ttuple', 'tunion', 'tliteral', 'tgeneric', 'tclassvar', 'eqtype',
           'eqlist', 'eqtuple', 'equnion', 'eqliteral', 'checknan', 'trychecks', 'tryiter', 'isstrbytes', 'notstrbytes',
           'isstrq', 'notiterstr', 'isallsame', 'aliascheck']

# %% ../nbs/02_grds.ipynb 6
import os, inspect
from functools import wraps

# %% ../nbs/02_grds.ipynb 8
from types import NoneType 
from typing import (
    Any, Set, Dict, List, Tuple, Union, ClassVar, Generic, Literal, 
    Callable, Iterable, get_args, get_origin, Type, TypeGuard,
)

# %% ../nbs/02_grds.ipynb 10
#| export


# %% ../nbs/02_grds.ipynb 12
#| export


# %% ../nbs/02_grds.ipynb 14
from .type import (T, GuardFunc, StrFactory,)
from .util import (guardname, guarddocs, guardanns,)

# %% ../nbs/02_grds.ipynb 17
def guard(
    *types: Type[T], 
    docsfn: StrFactory = guarddocs, 
    namefn: StrFactory = guardname, 
    annsfn: StrFactory = guardanns,
    negate: bool = False,
    optional: bool = False,
) -> Callable[[Callable[[Any], bool]], Callable[[Any], TypeGuard[T]]]:
    '''Creates a guard decorator based on the provided types and options.

    Parameters
    ----------
    types : Type[T]
        A variable number of types to be used in the guard.
    docsfn : StrFactory, optional
        Function to generate the docstring, by default guarddocs.
    namefn : StrFactory, optional
        Function to generate the name, by default guardname.
    annsfn : StrFactory, optional
        Function to generate the annotations, by default guardanns.
    negate : bool, optional
        Indicates if the guard is for negation, by default False.
    optional : bool, optional
        Indicates if the type can be optional (i.e., include None), by default False.

    Returns
    -------
    Callable[[Callable[[Any], bool]], Callable[[Any], TypeGuard[T]]]
        A decorator that applies the guard to a function.
    '''
    dtypes = (*types, ) + ((NoneType, ) if optional else ())
    def decorator(func:  Callable[[Any], bool]) -> Callable[[Any], TypeGuard[T]]:
        @wraps(func, updated=())
        def wrapper(x: Any) -> TypeGuard[T]:
            return not isinstance(x, dtypes) if negate == True else isinstance(x, dtypes)
        wrapper.__doc__ = docsfn(*dtypes)
        wrapper.__name__ = namefn(*dtypes)
        wrapper.__annotations__ = annsfn(*dtypes)
        return wrapper
    return decorator

# %% ../nbs/02_grds.ipynb 19
def tset(t: Type) -> bool:
    '''Checks if `t` type is a `Set`.'''
    return get_origin(t) is Set

def tlist(t: Type) -> bool: 
    '''Checks if `t` type is a `List`.'''
    return get_origin(t) is List

def tdict(t: Type) -> bool: 
    '''Checks if `t` type is a `Dict`.'''
    return get_origin(t) is Dict

def ttype(t: Type) -> bool: 
    '''Checks if `t` type is a `Type`.'''
    return get_origin(t) is type

def ttuple(t: Type) -> bool: 
    '''Checks if `t` type is a `Tuple`.'''
    return get_origin(t) is Tuple

def tunion(t: Type) -> bool: 
    '''Checks if `t` type is a `Union`.'''
    return get_origin(t) is Union

def tliteral(t: Type) -> bool: 
    '''Checks if `t` type is a `Literal`.'''
    return get_origin(t) is Literal

def tgeneric(t: Type) -> bool: 
    '''Checks if `t` type is a `Generic`.'''
    return get_origin(t) is Generic

def tclassvar(t: Type) -> bool: 
    '''Checks if `t` type is a `ClassVar`.'''
    return get_origin(t) is ClassVar


# %% ../nbs/02_grds.ipynb 21
def eqtype(val, obj: Type, guard: GuardFunc = tliteral) -> TypeGuard[Type]:
    '''Checks if `val` equals `obj`, `guard`'s typing type, or `obj`'s arguments if it is a typing type.

    Parameters
    ----------
    val : Any
        The value to be checked.

    typ : Type
        The type against which the value is to be checked.

    guard : GuardFunc, optional
        The guard function to apply, by default tliteral.

    Returns
    -------
    TypeGuard[Type]
        True if the value matches the specified type or its variants, False otherwise.
    '''
    if not guard(obj): return False
    try: return any((val == (args := get_args(obj)), val in args, val == obj))
    except: return False
    
def eqlist(val, obj: Type) -> TypeGuard[Type]: 
    '''Checks if `val` equals `obj` or any of obj's `List` arguments.'''
    return eqtype(val, obj, tlist)

def eqtuple(val, obj: Type) -> TypeGuard[Type]: 
    '''Checks if `val` equals `obj` or any of obj's `Tuple` arguments.'''
    return eqtype(val, obj, ttuple)

def equnion(val, obj: Type) -> TypeGuard[Type]: 
    '''Checks if `val` equals `obj` or any of obj's `Union` arguments.'''
    return eqtype(val, obj, tunion)

def eqliteral(val, obj: Type) -> TypeGuard[Type]: 
    '''Checks if `val` equals `obj` or any of obj's `Literal` arguments.'''
    return eqtype(val, obj, tliteral)


# %% ../nbs/02_grds.ipynb 23
from math import isnan, nan
from operator import eq

def checknan(val, obj: Type) -> TypeGuard[nan]:
    if isnan(obj): return isnan(val)
    return False

def trychecks(val, obj: Type) -> TypeGuard[Type]:
    fns = (isinstance, issubclass, eq)
    try: 
        if isnan(obj): fns += (checknan, )
    except: ...
    
    for f in fns:
        try: 
            if f(val, obj): return True            
        except: ...
    return False

def tryiter(val) -> TypeGuard[Iterable]:
    try:
        iter(val)
        return True
    except: return False

# %% ../nbs/02_grds.ipynb 24
@guard(str, bytes)
def isstrbytes(x): ...
@guard(str, bytes, negate=True)
def notstrbytes(x): ...
@guard(str, optional=True)
def isstrq(x): ...

def notiterstr(x) -> TypeGuard[Iterable['NotStrBytes']]:
    '''Checks if `x` is an iterable containing elements other than strings or bytes.'''
    return isinstance(x, Iterable) and notstrbytes(x)    

def isallsame(it: Iterable, dtype: T) -> TypeGuard[Iterable[T]]:
    '''Checks if all elements in an iterable are of the same specified type.'''
    itersafe = tryiter(it)
    if not itersafe: return trychecks(it, dtype)
    for el in it:
        if notiterstr(el):
            try: 
                if not isallsame(el, dtype): return False
            except TypeError: 
                return False            
        elif not trychecks(el, dtype):
            return False
    return True

# %% ../nbs/02_grds.ipynb 26
def aliascheck(obj, types, attrs: dict = dict()):
    '''Checks if an object matches a set of types and specific attributes.

    Parameters
    ----------
    obj : Any
        The object to be checked.
    types : iterable
        An iterable of types to check against the object.
    attrs : dict, optional
        A dictionary of attributes to check on the object, by default an empty dict.

    Returns
    -------
    bool
        True if the object matches the types and attributes, False otherwise.
    '''
    found = False
    for t in types:
        try:
            if (found := isinstance(obj, t)): break
        except TypeError:
            try:
                if (found := eqlist(obj, t)): break
                if (found := eqtuple(obj, t)): break
                if (found := equnion(obj, t)): break
                if (found := eqliteral(obj, t)): break
                if (found := obj == t): break
            except: ...

    if not found: return False
    
    for attr, val in attrs.items():        
        if getattr(obj, attr, None) == val: continue
        if attr in {'types', 'dtype', 'type'} and notiterstr(obj) and isallsame(obj, val): continue                     
        return False        
    return True



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/XX_tmpl.ipynb.

# %% auto 0
__all__ = []

# %% ../nbs/XX_tmpl.ipynb 6
#| export


# %% ../nbs/XX_tmpl.ipynb 8
from typing import List, Union, TypeAlias, Callable, Optional, ParamSpec, Type, Callable, Tuple

# %% ../nbs/XX_tmpl.ipynb 10
#| export


# %% ../nbs/XX_tmpl.ipynb 12
#| export


# %% ../nbs/XX_tmpl.ipynb 14
#| export


# %% ../nbs/XX_tmpl.ipynb 16
#| export




# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_ptyp.ipynb.

# %% auto 0
__all__ = ['StrBytes', 'NotStrBytes', 'StrQ', 'IntQ', 'notiterstr', 'isallsame']

# %% ../nbs/04_ptyp.ipynb 6
import os, inspect
from abc import abstractmethod
from functools import wraps, singledispatch

# %% ../nbs/04_ptyp.ipynb 8
from types import NoneType 
from typing import (Iterable, TypeGuard, )

# %% ../nbs/04_ptyp.ipynb 10
#| export


# %% ../nbs/04_ptyp.ipynb 12
#| export


# %% ../nbs/04_ptyp.ipynb 14
from .type import (T, )
from .util import (classname, qualname, settypes)
from .grds import (isallsame, aliascheck, trychecks, tryiter)
from .meta import (Alias, Not, Opt)

# %% ../nbs/04_ptyp.ipynb 16
@Alias(str, bytes)
class StrBytes: ...

@Not(str, bytes)
class NotStrBytes: ...

@Opt(str)
class StrQ: ...

@Opt(int)
class IntQ: ...

# %% ../nbs/04_ptyp.ipynb 17
def notiterstr(x) -> TypeGuard[Iterable[NotStrBytes]]:
    return isinstance(x, Iterable) and isinstance(x, NotStrBytes)

def isallsame(it: Iterable, dtype: type) -> TypeGuard[Iterable[T]]:
    itersafe = tryiter(it)
    if not itersafe: return trychecks(it, dtype)
    for el in it:
        if notiterstr(el) and not isallsame(el, dtype): return False
        elif not trychecks(el, dtype): return False
    return True
